// header file
#include "audioComponent.h"

// bell0bytes core
#include "app.h"

// bell0bytes audio
#include "XAudio2.h"

// bell0bytes util
#include "serviceLocator.h"

namespace audio
{
	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// Constructors and Destructor ///////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	AudioComponent::AudioComponent()
	{
		try { engine = new AudioEngine(); }
		catch (std::runtime_error& e) { throw e; }
	}

	AudioComponent::~AudioComponent()
	{
		if (engine)
			delete engine;
	}

	SoundEvent::SoundEvent() : sourceVoice(nullptr), waveFormat(nullptr)
	{
		audioData.clear();
		ZeroMemory(&audioBuffer, sizeof(XAUDIO2_BUFFER));
	}

	SoundEvent::~SoundEvent()
	{
		// clear the buffers
		audioData.clear();
		ZeroMemory(&audioBuffer, sizeof(XAUDIO2_BUFFER));

		// delete the wave format
		if(waveFormat)
			delete waveFormat;

		// delete the source voice
		if (sourceVoice)
		{
			sourceVoice->DestroyVoice();
			sourceVoice;
		}
	}
	

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////// Load Files //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> AudioComponent::loadFile(const std::wstring fileName, SoundEvent& soundEvent)
	{
		// handle errors
		util::Expected<void> result;
		HRESULT hr = S_OK;

		// load file into wave
		result = engine->loadFile(fileName, soundEvent.audioData, &soundEvent.waveFormat, soundEvent.waveLength);
		if (!result.isValid())
			return result;

		// create source voice
		hr = engine->dev->CreateSourceVoice(&soundEvent.sourceVoice, soundEvent.waveFormat);
		if (FAILED(hr))
			return std::runtime_error("Critical error: Unable to create source voice!");

		ZeroMemory(&soundEvent.audioBuffer, sizeof(XAUDIO2_BUFFER));
		soundEvent.audioBuffer.AudioBytes = (UINT32)soundEvent.audioData.size();
		soundEvent.audioBuffer.pAudioData = (BYTE* const)&soundEvent.audioData[0];
		soundEvent.audioBuffer.pContext = nullptr;

		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// Play //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> AudioComponent::playSoundEvent(const SoundEvent& audioEvent)
	{
		// handle errors
		HRESULT hr = S_OK;

		// submit the audio buffer to the source voice
		hr = audioEvent.sourceVoice->SubmitSourceBuffer(&audioEvent.audioBuffer);
		if (FAILED(hr))
			return std::runtime_error("Critical error: Unable to submit source buffer!");

		// start the source voice
		audioEvent.sourceVoice->Start();

		// return success
		return { };
	}

	util::Expected<void> AudioComponent::stopSoundEvent(const SoundEvent& audioEvent)
	{
		audioEvent.sourceVoice->Stop();

		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// Messages //////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> AudioComponent::onMessage(const core::Depesche& depesche)
	{
		if (depesche.type == core::DepescheTypes::PlaySoundEvent)
		{
			// handle errors
			HRESULT hr = S_OK;

			// submit the audio buffer to the source voice
			hr = ((SoundEvent*)depesche.message)->sourceVoice->SubmitSourceBuffer(&((SoundEvent*)depesche.message)->audioBuffer);
			if (FAILED(hr))
				return std::runtime_error("Critical error: Unable to submit source buffer!");

			// start the source voice
			((SoundEvent*)depesche.message)->sourceVoice->Start();
		}
		else if (depesche.type == core::DepescheTypes::StopSoundEvent)
		{
			((SoundEvent*)depesche.message)->sourceVoice->Stop();
		}

		// return success
		return { };
	}
}